;; Source code for "Learn to Solder" mystery hunt puzzle
	TITLE	"Mystery Hunt 2012"
	LIST	P=12F509, R=DEC
	#include <p12f509.inc>

	;; XXX: turn _CP_ON before deploying to teams
	;; XXX: turn on watchdog?
	__CONFIG _MCLRE_OFF & _CP_ON & _WDT_OFF & _IntRC_OSC

	;; useful PIC macros
	; The BTFSS and BTFSC instructions can be confusing to read.  It
        ; is easier to read the code with macros for IFSET and IFCLR
        ; meaning the opposite of SKPCLR and SKPSET respectively.

        #DEFINE IFSET   BTFSC
        #DEFINE IFCLR   BTFSS

        #DEFINE IFZ     SKPNZ
        #DEFINE IFNZ    SKPZ
        #DEFINE IFC     SKPNC
        #DEFINE IFNC    SKPC

        #DEFINE SKPSET  BTFSS
        #DEFINE SKPCLR  BTFSC

	;; dot duration (ms) = 1200 / WPM
	#DEFINE DOT_DUR_MS	230
	;; timing granuarity for morse input.  max duration=255*INPUT_TIME_MS
	#DEFINE INPUT_TIME_MS	10
	;; "Error" character we use in meta table ('*')
	#DEFINE ERROR_CHAR	0x2A
	;; turn on easter egg?
	#DEFINE INCLUDE_EASTER_EGG 1
	;; debug features
	#DEFINE STUFF_MEMORY	0
	#DEFINE DUMP_SYMBOL_TIMINGS 0
	#DEFINE INCLUDE_DEBUGW  0

	;; RAM locations
dlytmp	EQU	0x07		; temp used by delay routine and morse_fetch
orient	EQU	0x08		; bit flags, see definitions below
#define O_BIT_ORIENT           0
#define O_BIT_LED_COLOR        1
#define O_BIT_EASTER_EGG       2
#define O_BIT_DO_ABORT_CHECK   3
#define O_BIT_WAS_ZERO         4
#define O_BIT_SAW_ZERO         5
#define O_BIT_ABORT_DEBOUNCE   5 ; share space with SAW_ZERO
#define O_BIT_TEMP             7
mloc	EQU	0x09
keywait EQU	0x09		; share space with mloc
symmin	EQU	0x09		; share space with mloc/keywait
symmax	EQU	0x0A
oldmin	EQU	0x0B		; for clustering algorithm
oldmax	EQU	0x0C
avg_on	EQU	0x0B
hash	EQU	0x0A		; share space with symmax
hash2	EQU	0x0B		; share space with avg_on

dbgtmp1	EQU	0x0E
dbgtmp2	EQU	0x0F

mchar3	EQU	0x10		; msb (big-endian)
mchar2	EQU	0x11
mchar1	EQU	0x12
mchar0	EQU	0x13		; lsb

morse_in_start EQU 0x14		; 0x14-0x1F + 0x30-0x3F = 28 bytes

	; The PIC12C509 reset vector jumps to top of memory, then the
        ; program counter rolls to 0x00 after loading the RC osc.
        ; calibration value into W
	ORG	0x00 		; processor reset vector
	MOVWF	OSCCAL
	GOTO	main		; jump to program start

	;; JUMP TABLE for subroutines in high memory (0x100 or above)
morse_char_out			; trampoline for hi-memory subroutine
	GOTO	morse_char_out_HIGH
morse_mchar_out
	GOTO	morse_mchar_out_HIGH
do_word_delay
	GOTO	do_word_delay_HIGH
do_4dot_delay
	GOTO	do_4dot_delay_HIGH

	;; TRAMPOLINES for subroutines in really high memory (0x200 or above)
	ERRORLEVEL -306
	if INCLUDE_DEBUGW
debugW				;
	PAGESEL	debugW_HIGH	; really high trampoline
	GOTO	debugW_HIGH
	endif

	if INCLUDE_EASTER_EGG
show_easter_egg
	PAGESEL	show_easter_egg_HIGH
	GOTO	show_easter_egg_HIGH
	endif

	if STUFF_MEMORY
debug_stuff_memory
	PAGESEL	debug_stuff_memory_HIGH
	GOTO	debug_stuff_memory_HIGH
	endif

	ERRORLEVEL +306

	;;  TABLE DEREFERENCE ROUTINES

	;; lookup W in the 256-entry metatable
	;; (target address is 0x200 + W)
	;; need to clear PA0 on return
deref_high_W
	BSF	STATUS,PA0
	;; FALL THROUGH
derefW	;; dereference W (the target address will contain a RETLW)
	;; used for tables in 0x00-0xFF
	MOVWF	PCL

	;; ---------------------------------------------------------------
	;; start of program
	;; (this orientation check is kept in low 64 bytes to provide a
	;;  nice hint at the flip aha for those who do a disassembly)
main
	CLRF	STATUS
	MOVLW	B'10000000' 	; disable wakeup-on-pin, enable weak pull-up
				; (250uA); timer0=F_OSC/8 (prescaler, etc)
				; (timer0=500kHz)
	OPTION

	;;  clear RAM from 0x07 to 0x17 for consistent startup
	MOVLW	0x07
	MOVWF	FSR
clrloop
	CLRF	INDF
	INCF	FSR,F
	MOVF	FSR,W
	XORLW	0x18 | 0xC0	; remember MSBs of FSR read as '11'
	IFNZ
	 GOTO	clrloop		; if counter was not 0x18, then loop again
	CLRF	FSR		; ensure FSR:5 page bit is zero!

check_orientation
	MOVLW	B'11001111'	; set GP4/5 output, GP0/1 input, GP2/3 INPUT
	                        ; XXX to save power, make GP3 an output and
	                        ;     drive it low.
	TRIS	GPIO
	CLRF	GPIO		; GPIO outputs all 0
	MOVLW	1
	CALL	delayMS		; let everything settle

	;; orientation 1: GP0/GP5 is the LED
	;; orientation 2: GP1/GP4 is the LED
	IFCLR	GPIO, 0
	  GOTO	orient1
	IFCLR	GPIO, 1
	  GOTO	orient2
self	GOTO	self		; should never happen!

orient1				; GP0/GP5 is LED; GP1/GP4 is switch
	MOVLW	B'11001110'
	TRIS	GPIO
	CLRF	orient
	GOTO	main_loop
orient2				; GP1/GP4 is LED; GP0/GP5 is switch
	MOVLW	B'11001101'
	TRIS	GPIO
	MOVLW	(1<<O_BIT_ORIENT)|(1<<O_BIT_LED_COLOR)
	MOVWF	orient
	GOTO	main_loop

welcome
	;dt	"SOS 0123456789 * " ; test
	dt	"ENTER MORSE NO SPACES"
	dt	0

	;; Morse code
	#include "morse.inc"

	;; ------------------------------------------------
	;; subroutines
delayMS	; delay the # of milliseconds given in the W register.
	; (1ms = 1000 cycles)
	BCF	orient,O_BIT_ABORT_DEBOUNCE ; 1 cycle
anotherMS
	;;  this three instruction loop takes 3*256 = 768 cycles.
	CLRF	dlytmp		; 1 cycle
dloop1	INCFSZ	dlytmp,F	; 3*256-1 cycles
	GOTO	dloop1
	;; ok, we're at 768 cycles here.
	COMF	dlytmp,F
	BCF	dlytmp,6	; after this, dlytmp=1011 1111
	BCF	dlytmp,2	; after this, dlytmp=1011 1011
	BCF	dlytmp,1	; after this, dlytmp=1011 1001 = 185 (256-71)
	;; we're at 772 cycles here
dloop2	INCFSZ	dlytmp,F	; 71*3-1 cycles = 212 cycles
	GOTO	dloop2
	;; we're at 984 cycles here
	MOVWF	dlytmp		;ttl # of ms
	;; we're at 985 cycles here
	GOTO	$+1
	NOP
	;; we're at 988 cycles here; switch check will take 9
	MOVLW	B'10'
	IFSET	orient,O_BIT_ORIENT
	  MOVLW	B'01'
	ANDWF	GPIO, W		; sets Z flag if switch is pressed
	IFCLR	orient,O_BIT_DO_ABORT_CHECK
	  IORLW	1		; clear Z flag if we don't want to do abort
	IFNZ
	  GOTO	no_abort	; 997 cycles when taken
	;; DEBOUNCE switch.  Timing loop is slowed when switch pressed, but
	;; that's ok.
	IFSET	orient,O_BIT_ABORT_DEBOUNCE
	  ;; BAIL! jump directly to listen_for_morse, screw the call stack
	  GOTO	listen_for_morse
	BSF	orient,O_BIT_ABORT_DEBOUNCE
	;; fall through, wait another millisecond before checking again
no_abort
	;; we're at 997 cycles here
	DECFSZ	dlytmp,W
	  GOTO	anotherMS	; no_abort->delayMS = 3 cycles (taken)
	;; done!
	RETLW	0

green_led
	MOVLW   B'00000011'     ; GP1 high, GP5 low (keep GP4 low for button)
	MOVWF   GPIO
	RETLW	0
red_led
	CLRF	GPIO		; GP0/1 low, GP4/5 high (but keep other low)
	IFCLR	orient,O_BIT_ORIENT
	  BSF	GPIO, 5
	IFSET	orient,O_BIT_ORIENT
	  BSF	GPIO, 4
	RETLW	0
led_off
	CLRF	GPIO
	RETLW	0
led_on
	IFSET	orient,O_BIT_LED_COLOR
	  GOTO	green_led
	GOTO	red_led
toggle_led_color
	MOVLW	(1 << O_BIT_LED_COLOR)
	XORWF	orient,F
	RETLW	0

read_switch
	MOVLW	B'10'
	IFSET	orient,O_BIT_ORIENT
	  MOVLW	B'01'
	ANDWF	GPIO, W		; sets Z flag if switch is pressed
	RETLW	0

clear_mchar
	CLRF	mchar0
	CLRF	mchar1
	CLRF	mchar2
	CLRF	mchar3
	RETLW	0

find_minmax
	CLRF	symmin
	CLRF	symmax
	COMF	symmin,F
	MOVLW	morse_in_start
minmaxloop
	IORLW	0x10
	MOVWF	FSR
	;; check for zeros
	MOVF	INDF,W
	IFZ
	  GOTO	minmaxloop_done	; first zero terminates data
	;; check against oldmin; skip if less than or equal to oldmin
	SUBWF	oldmin,W	; oldmin - *INDF
	MOVF	INDF,W
	IFC			;for SUBWF, C=1->no borrow; C=0->borrow occurred
          GOTO	skip_for_min
	;; compare INDF to symmin
	SUBWF	symmin,W	; symmin - *INDF
	MOVF	INDF,W
	IFC			;for SUBWF, C=1->no borrow; C=0->borrow occurred
	  MOVWF	symmin
skip_for_min

	;; check against oldmax; skip if greater than oldmax (NOTE ASSYMMETRY)
	SUBWF	oldmax,W	; oldmax - *INDF
	MOVF	INDF,W
	IFNC			;for SUBWF, C=1->no borrow; C=0->borrow occurred
	  GOTO	skip_for_max
	SUBWF	symmax,W	; symmax - *INDF
	MOVF	INDF,W
	IFNC			;for SUBWF, C=1->no borrow; C=0->borrow occurred
	  MOVWF symmax
skip_for_max

	;; ok, now symmin and symmax have been updated
skip_for_minmax
	INCFSZ	FSR,W		; remember MSBs of FSR read as '11'
	  GOTO	minmaxloop
minmaxloop_done
	CLRF	FSR		; ensure that FSR:5 page bit is zero
	RETLW	0

	;; ---------------------------------------------------------------
	;; start of main loop
main_loop
	BSF	orient,O_BIT_DO_ABORT_CHECK
	CALL	led_off
	MOVLW	welcome
	MOVWF	mloc
	CALL	toggle_led_color

welcome_next_char
	;; load the next character
	MOVF	mloc,W
	CALL	derefW
	IORLW	0		; set the Z bit if W==0
	IFZ
	  GOTO	welcome_done
	CALL	morse_char_out
	INCF	mloc,F
	GOTO	welcome_next_char

welcome_done
	;; now check for key press for a while (2s)
	MOVLW	200
	MOVWF	keywait
welcome_done_loop
	DECF	keywait,F
	IFZ
	  GOTO	main_loop	; no dice, start over from top
	MOVLW	10
	CALL	delayMS		; w/ abort check to jump to listen_for_morse
				; if key pressed
	GOTO	welcome_done_loop

	;; it was, it was a key press!
listen_for_morse
	BCF	orient,O_BIT_DO_ABORT_CHECK ; turn off abort check!
	CALL	toggle_led_color
	;; initialize input area (0x14 - 0x1F; 0x30 - 0x3F )
	CALL	clear_mchar
parse_again
	;; reset pointer to start of area
	MOVLW	morse_in_start
	MOVWF	FSR
	CLRF	INDF
	CALL	led_on		; switch was pressed when we got here
	;; wait for the switch to be released
wait_for_release
	INCF	INDF,F
	IFZ
	  DECF	INDF,F		; saturating arithmetic
	CALL	read_switch	; set Z if switch is pressed
	IFNZ
	  GOTO	maybe_release
	MOVLW	INPUT_TIME_MS
	CALL	delayMS
	GOTO	wait_for_release
maybe_release
	MOVLW	INPUT_TIME_MS
	CALL	delayMS
	CALL	read_switch	; set Z if switch is pressed
	IFZ
	  GOTO	wait_for_release ; nope, just a transient
	;; okay, really released!
	CALL	led_off
	INCF	FSR,W
	IFZ
	  MOVLW 0x3F		; don't overflow, just overwrite last symbol
	IORLW	0x10		; ensure FSR jumps from 0x1F to 0x30
	MOVWF	FSR		; ok, we're pointing at next symbol now
	CLRF	INDF		; always zero-terminated (esp if we're stuck at 0x3F)

	;; wait for next press.
wait_for_press
	INCF	INDF,F
	IFZ
	  GOTO	end_of_word	; pause too long, assume word's over.
	CALL	read_switch	; set Z if switch is pressed
	IFZ
	  GOTO	maybe_press
	MOVLW	INPUT_TIME_MS
	CALL	delayMS
	GOTO	wait_for_press
maybe_press
	MOVLW	INPUT_TIME_MS
	CALL	delayMS
	CALL	read_switch	; set Z if switch is pressed
	IFNZ
	  GOTO	wait_for_press ; nope, just a transient
	;; okay, really pressed!
	CALL	led_on
	CLRF	INDF		; wipe out pause timing
	GOTO	wait_for_release

end_of_word
	CLRF	FSR		; ensure FSR:5 page bit is zero!

	if STUFF_MEMORY 	; DEBUG parse/hash/playback
	CALL	debug_stuff_memory
	endif

	;; okay!  We've captured a word!  We should process it and play it
	;; back.
	;; find min and max pause/symbol length
	CLRF	oldmin
	CLRF	oldmax
	COMF	oldmax,F
	CALL	find_minmax
	GOTO	gaps_are_same_size ; initialize w/ oldmin & oldmax = min/(max-1)
cluster_again
	;; if oldmin >= (oldmax+1), then we're done
	INCF	oldmax,W
	SUBWF	oldmin,W	; oldmin - oldmax
	IFC		   ; for SUBWF, C=1->no borrow; C=0->borrow occurred
	  GOTO	cluster_done
	CALL	find_minmax
	;; increase cluster cutoff to include the nearest thing
	;; |symmin - oldmin| vs |symmax - (oldmax+1)|
	;; (symmin - oldmin) vs ((oldmax+1) - symmax)
	MOVF	oldmin,W
	SUBWF	symmin,W   ; symmin - oldmin
	MOVWF	dlytmp
	DECF	symmax,W
	SUBWF	oldmax,W   ; (oldmax+1) - symmax
	SUBWF	dlytmp,W   ; (symmin-oldmin) - ((oldmax+1)-symmax)
	IFZ
	  GOTO	gaps_are_same_size
	IFC		   ; for SUBWF, C=1->no borrow; C=0->borrow occurred
	  GOTO	max_gap_is_smaller
min_gap_is_smaller
	;; make symmin the new oldmin
	MOVF	symmin,W
	MOVWF	oldmin
	GOTO	cluster_again
gaps_are_same_size
	MOVF	symmin,W
	MOVWF	oldmin
max_gap_is_smaller
	;; make (symmax-1) the new oldmax
	MOVLW	1
	SUBWF	symmax,W	;findminmax skips if *greater* than oldmax
	MOVWF	oldmax		; oldmax = symmax - 1
	GOTO	cluster_again

cluster_done
	if 1
	;; cutoff is oldmin - 1
	DECF	oldmin,W
	MOVWF	avg_on
	else
	;; compute average of oldmin/oldmax
	MOVF	oldmin,W
	ADDWF	oldmax,F
	RRF	oldmax,W	; now W = average
	MOVWF	avg_on
	endif

	;; now iterate through the data again, collecting symbol times
	MOVLW	32		;always loop 32 times
	MOVWF	dlytmp
	MOVLW	morse_in_start
	BCF	orient,O_BIT_SAW_ZERO
	BCF	orient,O_BIT_WAS_ZERO
parseloop
	IORLW	0x10
	MOVWF	FSR
	MOVF	INDF,W		; load the next duration
	IFZ
	  BSF	orient,O_BIT_SAW_ZERO
	BCF	orient,O_BIT_TEMP        ; save FSR:5
	IFSET	FSR,5
	  BSF	orient,O_BIT_TEMP
	BCF	FSR,5		; clear FSR:5 so we can safely access mchar0-3
	SUBWF	avg_on,W	; avg_on - *INDF. C=0->borrow occurred
	                        ; ie C = (avg_on >= *INDF)
	IFCLR	orient,O_BIT_SAW_ZERO ; 10...0 fill past end
	  GOTO	parse_notzero
	IFSET	orient,O_BIT_WAS_ZERO
	  BCF	STATUS,C
	IFCLR	orient,O_BIT_WAS_ZERO
	  BSF	STATUS,C
	BSF	orient,O_BIT_WAS_ZERO
parse_notzero
	RLF	mchar0,F	; move carry into LSB (big-endian)
	RLF	mchar1,F	; (1=dot, 0=dash)
	RLF	mchar2,F
	RLF	mchar3,F
	IFSET	orient,O_BIT_TEMP
	  BSF	FSR,5		; restore FSR:5
	INCF	FSR,W		; remember MSBs of FSR read as '11'
	DECFSZ	dlytmp,F
	  GOTO	parseloop	; always go through loop 32 times
	CLRF	FSR		; ensure FSR:5 page bit is zero

        ;;;  Look up word, set hash/hash2 (or leave ERROR_CHAR in hash)
	MOVLW	ERROR_CHAR
	MOVWF	hash
	if	INCLUDE_EASTER_EGG
	BCF	orient, O_BIT_EASTER_EGG
	endif
	CLRF	mloc		; steps through entries in meta_table
check_next_word
	MOVLW	mchar3
	MOVWF	FSR		; counts the 4 bytes of mchar3-0
check_next_byte
	MOVF	mloc,W
	CALL	deref_high_W
	BCF	STATUS,PA0
	;; does this match mchar?
	XORWF	INDF,W
	IFNZ
	  GOTO	not_this_word
	;; is there another byte of mchar?
	INCF	mloc,F
	INCF	FSR,F
	IFCLR	FSR,2
	  GOTO	check_next_byte
	GOTO	found_a_match	; yay!
not_this_word
	;; skip ahead to next entry
	MOVLW	7
	IORWF	mloc,W
	;; any words left? this is a hacky way to do it, based on the
	;; character which happens to be at the end of the table
	MOVWF	mloc
	CALL	deref_high_W
	BCF	STATUS,PA0
	XORLW	A'7'
	IFZ
	  GOTO	done_checking
	;; nope, more entries to look at
	INCF	mloc,F
	GOTO	check_next_word

found_a_match
	;; ok, this is a match! (and *mloc points to the first answer letter)
	IFSET	orient,O_BIT_ORIENT
	  INCF	mloc,F 		; use the "other" answer.
	;; load 'hash'
	MOVF	mloc,W
	CALL	deref_high_W
	BCF	STATUS,PA0
	MOVWF	hash
	;; increment mloc by two
	INCF	mloc,F
	INCF	mloc,F
	;; load 'hash2'
	MOVF	mloc,W
	CALL	deref_high_W
	BCF	STATUS,PA0
	MOVWF	hash2
	;; check for easter egg
	if	INCLUDE_EASTER_EGG
	XORLW	ERROR_CHAR
	IFZ
	  BSF	orient, O_BIT_EASTER_EGG
	endif
	;; done! (fall through)

done_checking
	CLRF	FSR		; hygiene

	if DUMP_SYMBOL_TIMINGS		; DUMP SYMBOL TIMINGS for debugging
debug_again
	MOVLW	morse_in_start
debug_more
	IORLW	0x10
	MOVWF	FSR
	CALL	debugW
	MOVF	INDF,W
	BCF	orient,O_BIT_TEMP ; save FSR:5
	IFSET	FSR,5
	  BSF	orient,O_BIT_TEMP
	BCF	FSR,5		; clear FSR:5 so we can safely access mchar0-5
	CALL	debugW
	IFSET	orient,O_BIT_TEMP
	  BSF	FSR,5		; restore FSR:5
	INCFSZ	FSR,W
	  GOTO debug_more

	MOVLW	250
	CALL	delayMS
	MOVLW	250
	CALL	delayMS
	MOVLW	250
	CALL	delayMS
	MOVLW	250
	CALL	delayMS
	GOTO	debug_again

	else			; ! DUMP_SYMBOL_TIMINGS

	;; play back the input word
	CALL	toggle_led_color; flip LED color prior to playback
	CALL	morse_mchar_out
	CALL	do_4dot_delay	; delay for an inter-word time
				; (minus the letter delay we've already taken)
	CALL	toggle_led_color

	if INCLUDE_EASTER_EGG
	IFSET	orient, O_BIT_EASTER_EGG
	  GOTO	show_easter_egg
	endif

	;; output the hash result
	MOVF	hash,W
	CALL	morse_char_out
	MOVLW	ERROR_CHAR
	XORWF	hash,W
	IFZ
	  GOTO	skip_second_char_out
	MOVF	hash2,W
	CALL	morse_char_out
skip_second_char_out
	;; okay, done!
	CALL	do_4dot_delay	; delay for an inter-word time
				; (minus the letter delay we've already taken)

done_listening
	;; turn on check for abort
	BSF	orient,O_BIT_DO_ABORT_CHECK
	;; wait a while more
	MOVLW	250
	CALL	delayMS
	MOVLW	250
	CALL	delayMS
	;; start again!
	GOTO	main_loop

	endif			; end of !DUMP_SYMBOL_TIMINGS


morse_char_out_HIGH		; output morse corresponding to ASCII in W
				; uses subroutines->can't be CALLed from a sub
	;; okay, now load the two morse bytes corresponding to this character
	MOVWF	dlytmp
	XORLW	0x20
	IFZ
	  GOTO	do_word_delay_HIGH
	;;  A-Z = 0b010x xxxx
	;;  a-z = 0b011x xxxx
	;;  0-9 = 0b0011 xxxx
	CALL	clear_mchar	; clear mchar0-mchar3
	MOVLW	0x21
	SUBWF	dlytmp,F	; dlytmp - 0x21, C=1->no borrow (ie, dlytmp >= 0x21)
	IFNC
	  CLRF	dlytmp		; be safe, map all lower chars to !
	MOVLW	morse_punct
	ADDWF	dlytmp,W
	CALL	derefW		; look up this entry in the table
	MOVWF	mchar3		; note that mchar2, mchar1, mchar0 are already clear
	;; fall through to morse_mchar_out

morse_mchar_out_HIGH		;output morse corresponding to mchar0-3
	;;
next_bit
	;; turn LED on
	CALL	led_on
	;; delay one dot time
	MOVLW	DOT_DUR_MS
	CALL	delayMS
	;; put next bit into carry
	BCF	STATUS,C
	RLF	mchar0,F
	RLF	mchar1,F
	RLF	mchar2,F
	RLF	mchar3,F
	IFC
	  GOTO	skip_out_dash
	;; make this a dash, not a dot
	MOVLW	DOT_DUR_MS
	CALL	delayMS
	MOVLW	DOT_DUR_MS
	CALL	delayMS
skip_out_dash
	;; turn off LED
	CALL	led_off
	MOVLW	DOT_DUR_MS
	CALL	delayMS
	;; was this the last bit?
	MOVLW	0x7F
	ANDWF	mchar3,W
	IORWF	mchar2,W
	IORWF	mchar1,W
	IORWF	mchar0,W
	IFNZ
	  GOTO	next_bit
	;;  finish up
	GOTO	do_2dot_delay 	; letter delay minus the dot delay
				; we've already taken

do_word_delay_HIGH
	;; inter-word delay (7 bit times)
	MOVLW	DOT_DUR_MS
	CALL	delayMS
	MOVLW	DOT_DUR_MS
	CALL	delayMS
	MOVLW	DOT_DUR_MS
	CALL	delayMS
do_4dot_delay_HIGH
	MOVLW	DOT_DUR_MS
	CALL	delayMS
do_letter_delay
	;; inter-letter delay (3 bit times)
	MOVLW	DOT_DUR_MS
	CALL	delayMS
do_2dot_delay
	MOVLW	DOT_DUR_MS
	CALL	delayMS
	MOVLW	DOT_DUR_MS
	CALL	delayMS

	;; done with this letter
	RETLW	0

	;; LOOKUP TABLE FOR META MAPPING
	ORG	0x200
metatable
	#include "boring.inc"
metatable_end

	;;  HIGH MEMORY SUBROUTINES (be careful about GOTOs!)
	ERRORLEVEL -306

	if INCLUDE_EASTER_EGG
	;; easter egg routines
	;; whoo -- show the easter egg
show_easter_egg_HIGH
	MOVLW	LOW(easter_egg)
	MOVWF	mloc
easter_egg_next_char
	PAGESEL	show_easter_egg	; restore PA0 so we can call low-memory subs
	;; load the next character
	MOVF	mloc,W
	CALL	deref_high_W
	PAGESEL	show_easter_egg	; restore PA0 after high memory lookup
	MOVWF	dlytmp
	MOVLW	LOW(easter_egg - EASTER_EGG_KEY_OFFSET)
	SUBWF	mloc,W
	CALL	deref_high_W
	PAGESEL	show_easter_egg	; restore PA0 after high memory lookup
	XORWF	dlytmp,W	; DECRYPT!
	IFZ
	  GOTO	skip_second_char_out ; back to main loop
	CALL	morse_char_out
	INCF	mloc,F
	PAGESEL	easter_egg_next_char
	GOTO	easter_egg_next_char

	endif

	if INCLUDE_DEBUGW
debugW_HIGH			; flash W out on the LED
	PAGESEL	led_off		; allow calling low-memory subroutines
	MOVWF	dbgtmp1
	CALL	led_off
	MOVLW	250
	CALL	delayMS
	MOVLW	250
	CALL	delayMS
	CALL	green_led	; set the '0'
	MOVLW	9
	MOVWF	dbgtmp2
	PAGESEL	debug_turn_off
	GOTO	debug_turn_off
debug_next_bit
	PAGESEL	green_led	; restore after GOTO to debug_next_bit
	CALL	green_led
	BCF	STATUS,C
	RLF	dbgtmp1,F
	IFC
	  CALL	red_led
debug_turn_off
	PAGESEL	delayMS	; restore after GOTO to debug_turn_off
	MOVLW	250
	CALL	delayMS
	MOVLW	250
	CALL	delayMS
	CALL	led_off
	MOVLW	150
	CALL	delayMS
	PAGESEL debug_next_bit	; setup for potential GOTO to debug_next_bit
	DECFSZ	dbgtmp2,F
	  GOTO	debug_next_bit

	PAGESEL	delayMS		; restore PA0 for caller
	MOVLW	250		; quiet buffer at end
	CALL	delayMS
	RETLW	0
	endif

	;; Stuff memory with 'SCOTT' in morse, to test parsing routines.
	if STUFF_MEMORY
debug_stuff_memory_HIGH
	PAGESEL	debug_stuff_memory ; reset PA bit
	CLRF	FSR
	MOVLW	0x18
	MOVWF	(morse_in_start+0)|0x10
	MOVLW	0x15
	MOVWF	(morse_in_start+1)|0x10
	MOVLW	0x17
	MOVWF	(morse_in_start+2)|0x10
	MOVLW	0x41
	MOVWF	(morse_in_start+3)|0x10
	MOVLW	0x0C
	MOVWF	(morse_in_start+4)|0x10
	MOVLW	0x54
	MOVWF	(morse_in_start+5)|0x10
	MOVLW	0x0E
	MOVWF	(morse_in_start+6)|0x10
	MOVLW	0x35
	MOVWF	(morse_in_start+7)|0x10
	MOVLW	0x3B
	MOVWF	(morse_in_start+8)|0x10
	MOVLW	0x3B
	MOVWF	(morse_in_start+9)|0x10
	MOVLW	0x4C
	MOVWF	(morse_in_start+10)|0x10
	MOVLW	0x4D
	MOVWF	(morse_in_start+11)|0x10

	BSF	FSR,5
	ERRORLEVEL -302
	CLRF	(morse_in_start+12)|0x10
	CLRF	FSR
	RETLW	0

	endif

	END
